#define G_LOG_DOMAIN "gpod-manager"

#include "config.h"

#include "gpod-manager.h"

#include <tracker-sparql.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>

/**
 * SECTION:gpod-manager
 * @title: GpodManager
 * @short_description: manages feeds and saving of information to and from the
 *                     tracker database.
 * @include: gpod-manager.h
 *
 * The #GpodManager is object to manage the fetching, saving, and retrevial of
 * feeds and their items.
 **/

/*
// View tag info
tracker sparql -t mfo:FeedChannel -p

// Start all daemons, RSS miner may be off
tracker daemon --start
// Check miner status
tracker daemon

// Clear old Feeds & items
tracker sparql -u -q "DELETE { ?u a rdfs:Resource } WHERE { ?u a mfo:FeedChannel }"
tracker sparql -u -q "DELETE { ?u a rdfs:Resource } WHERE { ?u a mfo:FeedMessage }"

// Add new feeds / 30s check / <url>
tracker sparql -u -q "INSERT { _:FeedSettings a mfo:FeedSettings ; mfo:updateInterval 30 . _:Feed a nie:DataObject, mfo:FeedChannel ; mfo:feedSettings _:FeedSettings ; nie:url <http://h3h3roost.libsyn.com/rss> }"

// Get Feeds
tracker sparql -q "SELECT ?feed nie:title(?feed) WHERE { ?feed a mfo:FeedChannel }"
tracker sparql -q "SELECT ?feed nie:title(?feed) nie:description(?feed) mfo:image(?feed) nie:copyright(?feed) nmo:lastMessageDate(?feed) WHERE { ?feed a mfo:FeedChannel }"

// Wait a moment as feeds populate

tracker search --feeds

// Feeds should display

// Get via query
tracker sparql -q "SELECT ?feed nie:title(?feed) WHERE { ?feed a mfo:FeedMessage } ORDER BY ASC(nie:title(?feed))"
tracker sparql -q "SELECT ?feed nie:title(?feed) WHERE { ?feed a mfo:FeedMessage } ORDER BY ASC(nie:contentCreated(?feed))"
tracker sparql -q "SELECT ?feed nie:title(?feed) WHERE { ?feed a mfo:FeedMessage ; nie:url <http://h3h3roost.libsyn.com/rss> } ORDER BY ASC(nie:contentCreated(?feed))"
tracker sparql -q "SELECT ?feed nie:title(?feed) WHERE { ?feed a mfo:FeedMessage ; nmo:communicationChannel <urn:uuid:f6203b61-7c06-64c8-ef83-2a960e30525d> } ORDER BY DESC(nie:contentCreated(?feed)) LIMIT 10"

*/

typedef struct _GpodManagerPrivate
{
    TrackerSparqlConnection *tracker;
    GCancellable *query_cancellable;
} GpodManagerPrivate;

struct _GpodManager
{
    GObject parent;
};

G_DEFINE_TYPE_WITH_PRIVATE (GpodManager, gpod_manager, G_TYPE_OBJECT)

static void
add_feed_async_cb (GObject      *object,
                   GAsyncResult *result,
                   gpointer      user_data)
{
    g_return_if_fail (TRACKER_SPARQL_IS_CONNECTION (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));

    TrackerSparqlConnection *connection = TRACKER_SPARQL_CONNECTION (object);
    GError *error = NULL;

    tracker_sparql_connection_update_finish (connection, result, &error);

    if (error != NULL)
    {
        g_warning ("Unable to add feed to Tracker database: %s", error->message);
        g_error_free (error);
        return;
    }
}

static void
add_feed_async (GpodManager  *self,
                gchar        *feed_url,
                GCancellable *cancellable)
{
    g_return_if_fail (GPOD_IS_MANAGER (self));
    g_return_if_fail (feed_url != NULL);
    g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));

    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);
    GString *query = g_string_new (NULL);

    g_string_printf (query,
                     "INSERT { "
                     "    _:FeedSettings a mfo:FeedSettings ; "
                     "        mfo:updateInterval %i . "
                     "    _:Feed a nie:DataObject, mfo:FeedChannel ; "
                     "        mfo:feedSettings _:FeedSettings ; "
                     "        nie:url <%s> "
                     "}",
                     30, feed_url); // 30 minutes

    tracker_sparql_connection_update_async (priv->tracker,
                                            query->str,
                                            G_PRIORITY_DEFAULT,
                                            cancellable,
                                            add_feed_async_cb,
                                            NULL);

    g_string_free (query, TRUE);
}

// TODO: Add GError
static void
get_feed (GpodManager  *self,
          gchar        *feed_url,
          GCancellable *cancellable)
{
    g_return_if_fail (GPOD_IS_MANAGER (self));
    g_return_if_fail (feed_url != NULL);
    g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));

    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);
    GString *query = g_string_new (NULL);
    GError *error = NULL;
    TrackerSparqlCursor *cursor;

    g_string_printf (query,
                     "SELECT "
                     "    ?urn AS urn"
                     "    ?title AS title "
                     "    ?description AS description "
                     "    ?copyright AS copyright "
                     "    ?updated AS updated "
                     "    ?image AS image "
                     "    ?updateInterval AS updateInterval "
                     "WHERE { "
                     "    ?urn a mfo:FeedChannel ; "
                     "        nie:url <%s> ; "
                     "        nie:title ?title ; "
                     "        nie:description ?description ; "
                     "        nie:copyright ?copyright ; "
                     "        nmo:lastMessageDate ?updated ; "
                     "        mfo:image ?image . "
                     "    OPTIONAL { "
                     "        ?urnSettings a mfo:FeedSettings ; "
                     "            mfo:updateInterval ?updateInterval . "
                     "    } "
                     "} "
                     "LIMIT 1", feed_url);

    g_print ("-------------------------------------------------\n");
    g_print ("Feed for: %s\n", feed_url);
    g_print ("-------------------------------------------------\n");
    g_print ("%s\n", query->str);
    g_print ("-------------------------------------------------\n");

    cursor = tracker_sparql_connection_query (priv->tracker, query->str, cancellable, &error);

    g_string_free (query, TRUE);

    if (error != NULL)
    {
        g_warning ("Unable to get feed from Tracker database: %s", error->message);
        g_error_free (error);
        return;
    }

    if (!cursor)
    {
        g_print ("%s\n", "No results were found");
        return;
    }

    g_print ("-------------------------------------------------\n");
    g_print ("Feed Info (%i):\n", tracker_sparql_cursor_get_n_columns (cursor));
    g_print ("-------------------------------------------------\n");

    while (tracker_sparql_cursor_next (cursor, NULL, NULL))
    {
        g_print ("Result:\n    %s\n    %s\n    %s\n    %s\n    %s\n    %s\n    %s\n\n",
                 tracker_sparql_cursor_get_string (cursor, 0, NULL),
                 tracker_sparql_cursor_get_string (cursor, 1, NULL),
                 tracker_sparql_cursor_get_string (cursor, 2, NULL),
                 tracker_sparql_cursor_get_string (cursor, 3, NULL),
                 tracker_sparql_cursor_get_string (cursor, 4, NULL),
                 tracker_sparql_cursor_get_string (cursor, 5, NULL),
                 tracker_sparql_cursor_get_string (cursor, 6, NULL));
        g_print ("-------------------------------------------------\n");
    }

    // TODO: Create GpodFeed struct wrapper type
    // GpodFeed *feed = g_new (GpodFeed, 1);

    g_object_unref (cursor);
}

/*
 *     while (tracker_sparql_cursor_next (cursor, NULL, NULL))
        g_print ("Feed: %s\n", tracker_sparql_cursor_get_string (cursor, 0, NULL));
 */





/*
static void
get_feed_async_cb (GObject      *object,
                   GAsyncResult *result,
                   gpointer      user_data)
{
    g_return_if_fail (TRACKER_SPARQL_IS_CONNECTION (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));

    TrackerSparqlConnection *connection = TRACKER_SPARQL_CONNECTION (object);
    GError *error = NULL;

    tracker_sparql_connection_query_finish (connection, result, &error);

    if (error != NULL)
    {
        g_warning ("Unable to get feed from Tracker database: %s", error->message);
        g_error_free (error);
        return;
    }
    else
    {
        tracker_sparql_cursor_next_async (cursor,
                                          query_data->cancellable,
                                          on_cursor_cb,
                                          query_data);
    }
}

static void
get_feed_async (GpodManager  *self,
                gchar        *feed_url,
                GCancellable *cancellable)
{
    g_return_if_fail (GPOD_IS_MANAGER (self));
    g_return_if_fail (feed_url != NULL);
    g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));

    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);
    GString *query = g_string_new (NULL);

    g_string_printf (query,
                     "SELECT "
                     "nie:title(?feed) "
                     "WHERE { "
                     "    ?feed a mfo:FeedMessage ; "
                     "        nie:url <%s>"
                     "}", feed_url);

    tracker_sparql_connection_query_async (priv->tracker,
                                           query->str,
                                           cancellable,
                                           get_feed_async_cb,
                                           NULL);

    g_string_free (query, TRUE);
}
*/
/*

static void
on_cursor_cb (GObject      *object,
              GAsyncResult *result,
              gpointer      user_data)
{

}

static void
get_feeds_async_cb (GObject      *object,
                    GAsyncResult *result,
                    gpointer      user_data)
{
    TrackerSparqlConnection *connection;
    TrackerSparqlCursor *cursor;
    GError *error = NULL;
    FeedGetQueryData *query_data;

    g_return_if_fail (TRACKER_SPARQL_IS_CONNECTION (object));

    query_data = user_data;

    cursor = tracker_sparql_connection_query_finish (connection,
                                                     result,
                                                     &error);

    if (error != NULL)
    {
        g_warning ("Error on fetching feeds: %s", error->message);

        //gpod_feed_view_add ();
        g_free (query_data);
    }
    else
    {
        tracker_sparql_cursor_next_async (cursor,
                                          query_data->cancellable,
                                          on_cursor_cb,
                                          query_data);
    }
}

static void
get_feeds_async (GpodManager  *self,
                 GList        *list,
                 GCancellable *cancellable)
{
    g_return_if_fail (GPOD_IS_MANAGER (self));

    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);
    GString *query = NULL;
    FeedGetQueryData *query_data = NULL;

    g_string_printf (query,
                     "SELECT "
                     "nie:url(?feed)"
                     "nie:title(?feed) "
                     "WHERE { "
                     "    ?feed a mfo:FeedChannel "
                     "} "
                     "ORDER BY nie:title(?feed)");

    query_data = g_new (FeedGetQueryData, 1);
    query_data->feeds = list;
    query_data->cancellable = cancellable;

    tracker_sparql_connection_query_async (priv->tracker,
                                           query->str,
                                           query_data->cancellable,
                                           get_feeds_async_cb,
                                           query_data);

    g_string_free (query, TRUE);
}
*/
/*
 * gpod_manager_finalize:
 * @object: An instance of #GpodManager
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
gpod_manager_finalize (GObject *object)
{
    GpodManager *self = GPOD_MANAGER (object);
    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);

    g_clear_object (&priv->tracker);

    g_cancellable_cancel (priv->query_cancellable);
    g_clear_object (&priv->query_cancellable);

    G_OBJECT_CLASS (gpod_manager_parent_class)->finalize(object);
}

/*
 * gpod_manager_init:
 * @self: An instance of #GpodManager
 *
 * Initializes various aspects of the #GpodManager. Anything the #GpodManager
 * would need before becoming presented should be done here.
 */
static void
gpod_manager_init (GpodManager *self)
{
    GpodManagerPrivate *priv = gpod_manager_get_instance_private (self);
    GError *error;

    priv->tracker = tracker_sparql_connection_get (NULL, &error);

    if (!priv->tracker)
    {
        if (error)
        {
            g_warning ("Error connecting to tracker: %s", error->message);
            g_error_free (error);
        }

        return;
    }

    priv->query_cancellable = g_cancellable_new ();
}

/*
 * gpod_manager_class_init:
 * @klass: An instance of #GpodManagerClass
 *
 * Initializes the various virtural functions in #GpodManagerClass. In
 * addtition it connects the various #GtkWidgets from the gresource file using
 * gtk_widget_class_set_template_from_resource().
 */
static void
gpod_manager_class_init (GpodManagerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = gpod_manager_finalize;
}

/**
 * gpod_manager_new:
 *
 * Creates a new #GpodManager, a custom widget to show all feeds user is
 * subscribed to.
 *
 * ## Usage
 * |[
 * GpodManager *manager;
 *
 * manager = gpod_manager_new ();
 * ]|
 *
 * Returns: #GpodManager
 */
GpodManager *
gpod_manager_new ()
{
    return g_object_new (GPOD_TYPE_MANAGER,
                         NULL);
}

void
gpod_manager_add_feed_by_url (GpodManager *self,
                              gchar       *feed_url)
{
    add_feed_async (self, feed_url, NULL);
}

void
gpod_manager_get_feed_by_url (GpodManager *self,
                              gchar       *feed_url)
{
    get_feed (self, feed_url, NULL);
}


GList *
gpod_manager_get_feeds (GpodManager *self)
{
    GList *feeds = NULL;

    return feeds;
}

GList *
gpod_manager_get_feed_items (GpodManager *self)
{
    GList *feeds = NULL;

    return feeds;
}
